
// Include file that defines a mesh2 object in the form of a human figure.
// Generated from the MakeHuman application using the export functionality.
// 
// This file is licensed under the terms of the CC-LGPL. 
// This license permits you to use, modify and redistribute the content.
// 
// Typical render time 1 minute (at 1280x1024 AA 0.3).
// Size is about 16 POV-Ray units high.
// Position is centred on the origin.      
//
// This include file contains 3 models that can be rendered separately
// or in any combination. The models and the options available are 
// detailed in the file MakeHuman.html.  
//

#include "math.inc"
